/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef ASSEMBLYJOBHELPERSTRATEGIES_CPP
#define ASSEMBLYJOBHELPERSTRATEGIES_CPP


#include "AssemblyJobHelperStrategies.h"

#include "AssemblyException.h"
#include "ScoredSeqShallow.h"
#include <iostream>
#include <limits.h>
using namespace::std;



float AssemblyJobHelperStrategies::_SCORE_MAX = float(INT_MAX / 2);



AssemblyJobHelperStrategies::ScoreCalculator::~ScoreCalculator(){}

AssemblyJobHelperStrategies::ScoreCalcAdd::ScoreCalcAdd(){}
AssemblyJobHelperStrategies::ScoreCalcAdd::~ScoreCalcAdd(){}


float AssemblyJobHelperStrategies::ScoreCalcAdd::adjustA(float a){ return a; }
float AssemblyJobHelperStrategies::ScoreCalcAdd::adjustB(float b){ return b; }
float AssemblyJobHelperStrategies::ScoreCalcAdd::aPlusB(float a, float b){
  if ( _SCORE_MAX - a < b ){ return _SCORE_MAX; }
  else { return a + b; }
}
float AssemblyJobHelperStrategies::ScoreCalcAdd::aMinusB(float a, float b){
  float diff = a - b;
  if ( diff < 0 ){ return 0.0;}
  else { return diff; }
}
float AssemblyJobHelperStrategies::ScoreCalcAdd::bMinusA(float b, float a){
  float diff = b - a;
  if ( diff < 0 ){ return 0.0;}
  else { return diff; }
}



AssemblyJobHelperStrategies::ScoreCalcNormalizeA::ScoreCalcNormalizeA(){};
AssemblyJobHelperStrategies::ScoreCalcNormalizeA::~ScoreCalcNormalizeA(){};

AssemblyJobHelperStrategies::ScoreCalcNormalizeA::ScoreCalcNormalizeA(int denominator) :
  _denominator(denominator){}
float AssemblyJobHelperStrategies::ScoreCalcNormalizeA::adjustA(float a){ return a / _denominator; }
float AssemblyJobHelperStrategies::ScoreCalcNormalizeA::adjustB(float b){ return b; }
float AssemblyJobHelperStrategies::ScoreCalcNormalizeA::aPlusB(float a, float b){
  a = adjustA(a);
  if ( _SCORE_MAX - a < b ){ return _SCORE_MAX; }
  else { return a + b; }
}
float AssemblyJobHelperStrategies::ScoreCalcNormalizeA::aMinusB(float a, float b){
  a = adjustA(a);
  float diff = a - b;
  if ( diff < 0 ){ return 0;}
  else { return diff; }
}
float AssemblyJobHelperStrategies::ScoreCalcNormalizeA::bMinusA(float b, float a){
  a = adjustA(a);
  float diff = b - a;
  if ( diff < 0 ){ return 0;}
  else { return diff; }
}

AssemblyJobHelperStrategies::ScoreCalcXy::ScoreCalcXy(){};
AssemblyJobHelperStrategies::ScoreCalcXy::~ScoreCalcXy(){};

AssemblyJobHelperStrategies::ScoreCalcXy::ScoreCalcXy(ScoreCalculator* calc, bool xIsA) :
  _calc(calc),
  _xIsA(xIsA){}
float AssemblyJobHelperStrategies::ScoreCalcXy::adjustX(float x){ 
  if ( _xIsA ){ return _calc->adjustA(x); }
  else { return _calc->adjustB(x); }
}
float AssemblyJobHelperStrategies::ScoreCalcXy::adjustY(float y){
  if ( _xIsA ){ return _calc->adjustB(y); }
  else { return _calc->adjustA(y); }
}
float AssemblyJobHelperStrategies::ScoreCalcXy::xPlusY(float x, float y){
  if ( _xIsA ){ return _calc->aPlusB(x,y); }
  else { return _calc->aPlusB(y,x); }
}
float AssemblyJobHelperStrategies::ScoreCalcXy::xMinusY(float x, float y){
  if ( _xIsA ){ return _calc->aMinusB(x,y); }
  else { return _calc->aMinusB(y,x); }
}
float AssemblyJobHelperStrategies::ScoreCalcXy::yMinusX(float y, float x){
  if ( _xIsA ){ return _calc->bMinusA(y,x); }
  else { return _calc->bMinusA(x,y); }
}




ScoredSeq* AssemblyJobHelperStrategies::combine(Alignment* alignment){
  ScoreCalculator* calc = new ScoreCalcAdd();
  ScoredSeq* outputSeq =  combine(alignment, calc);
  delete calc;
  return outputSeq;
}


ScoredSeq* AssemblyJobHelperStrategies::combine(Alignment* alignment, int denominator){
  ScoreCalculator* calc = new ScoreCalcNormalizeA(denominator);
  ScoredSeq* outputSeq =  combine(alignment, calc);
  delete calc;
  return outputSeq;
}





ScoredSeq* AssemblyJobHelperStrategies::combine(Alignment* alignment, ScoreCalculator* calc){
  if (! alignment->isLocked()){
    throw AssemblyException::ArgError("in combine, unlocked alignment should not be combined");
  }

  // these variables are defined just for ease of access
  ScoredSeq * seqA = alignment->seqA();
  ScoredSeq * seqB = alignment->seqB();
  char senseA = '+';
  char senseB = alignment->orientation();

  // the values for the constructor of the new ScoredSeq
  CombineDataCarrier * carrier = new CombineDataCarrier(seqA->size() + seqB->size());
  //carrier->newSeq = "";

  // establish where the beginning and end of each alignment are
  long posA = 0;
  long posB = 0;
  long sizeA = seqA->size();
  long sizeB = seqB->size();
  bool lastPosWasMatch = false;

  // make the decisions for the aligned segment
  while ( posA < sizeA or posB < sizeB ){

    if ( posA==sizeA or alignment->isGapped(posA,seqA) or posB==sizeB or alignment->isGapped(posB,seqB) ){
      long aStartPos = posA-1;
      long bStartPos = posB-1;
      // bring both positions to the point after the alignment block
      while ( posA < sizeA and alignment->isGapped(posA,seqA) ){ ++posA; }
      while ( posB < sizeB and alignment->isGapped(posB,seqB) ){ ++posB; }

      // check that block is legit
      if ( posA!=sizeA and alignment->isLinked(posA,seqA) and alignment->getLinkage(posA,seqA)==posB ){} // this is legit
      else if ( posA==sizeA and posB==sizeB ){} // this is also legit
      else {
	throw AssemblyException::LogicError("in combine, conditions after a gap block are not legit.");
      }

      // if a linkage was already added to the end, it must be removed so it can be replaced
      if (lastPosWasMatch){ carrier->decrementLink(); }
      combineGapBlock(alignment, aStartPos, posA, bStartPos, posB, carrier, calc);
      lastPosWasMatch = false;

    } else { // MATCH
      char nucA = seqA->nucAtPosition(posA,senseA);
      char nucB = seqB->nucAtPosition(posB,senseB);
      float scoreA = seqA->scoreAtPosition(posA,senseA);
      float scoreB = seqB->scoreAtPosition(posB,senseB);

      float linkA = 0.0;
      float linkB = 0.0;
      // neither, one or both of these may be the case; a zero value will be added if neither is
      // the case, which is fine since it will not be used for ScoredSeq construction
      if ( posA + 1 < sizeA ){ linkA = seqA->linkAfterPosition(posA,senseA); }
      if ( posB + 1 < sizeB ){ linkB = seqB->linkAfterPosition(posB,senseB); }
      float localLinkScore = calc->aPlusB(linkA,linkB);

      if ( nucA == nucB ){
	carrier->addNuc(nucA, calc->aPlusB(scoreA,scoreB), localLinkScore);
      } else if ( calc->adjustA(scoreA) >= calc->adjustB(scoreB) ){ // A is better or the same
	if ( nucA == 'N' ){
	  carrier->addNuc(nucB, 0.0, localLinkScore);
	} else {
	  carrier->addNuc(nucA, calc->aMinusB(scoreA,scoreB), localLinkScore);
	}
      } else { // here, B is better
	carrier->addNuc(nucB, calc->bMinusA(scoreB,scoreA), localLinkScore);
      }

      posA++;
      posB++;
      lastPosWasMatch = true;
    }
  }

  ScoredSeq * returnVal = carrier->getSeq();
  delete carrier;
  return returnVal;
}



// modifies "links"; this is a helper for combineGapBlock
void AssemblyJobHelperStrategies::getSegmentsFromBlock(map<long,float>* links, long startPos, long endPos, char sense,
						       ScoredSeq* seq, Alignment* alignment, ScoreCalcXy* calc){
  long pos = startPos;
  while (pos < endPos){
    if (pos==-1 or pos+1==seq->size()){ links->insert( pair<long,float> ( pos, calc->adjustX(0) ) ); }
    else if ( pos==startPos or pos+1==endPos or 
	      alignment->gapPairedAfter(pos,seq) != alignment->gapPairedAfter(pos+1,seq) ){
      links->insert( pair<long,float> ( pos, calc->adjustX( seq->linkAfterPosition(pos,sense) ) ) );
    }
    ++pos;
  }
}


// modifies "carrier"; this is a helper for combineGapBlock
void AssemblyJobHelperStrategies::fillInGaps(CombineDataCarrier* carrier,
					     ScoredSeq* seq, long startPos, long endPos, char sense,
					     float penaltyConstant, map<long,float>* links, ScoreCalcXy* calc){
  for (long n = startPos; n < endPos; ++n){
    // fill in non-link values
    if ( n != startPos ){
      carrier->addNuc(seq->nucAtPosition(n,sense), calc->adjustX( seq->scoreAtPosition(n,sense) ));
    }
    // the before and after links are imaginary and don't get added to the carrier
    if (n != -1 and n+1 != seq->size()){
      map<long,float>::iterator compScoreIt = links->find(n);
      if (compScoreIt==links->end()){ // this was a non-competitive position
	carrier->addLink( calc->adjustX( seq->linkAfterPosition(n,sense) ) );
      } else { // this was a competitive position, so the value is modified by the constant
	carrier->addLink( compScoreIt->second * penaltyConstant );
      }
    }
  }
}

void AssemblyJobHelperStrategies::combineGapBlock(Alignment* alignment,
						  long aStartPos, long aEndPos,
						  long bStartPos, long bEndPos,
						  CombineDataCarrier* carrier, ScoreCalculator* calc){

  ScoredSeq* seqA = alignment->seqA();
  ScoredSeq* seqB = alignment->seqB();
  char senseA = '+';
  char senseB = alignment->orientation();


  // collect the relevant (competing) link scores by defining gap segments within the block
  map<long,float> aLinks;
  map<long,float> bLinks;
  ScoreCalcXy * tempCalcA = new ScoreCalcXy(calc,true);
  ScoreCalcXy * tempCalcB = new ScoreCalcXy(calc,false);
  getSegmentsFromBlock(&aLinks,aStartPos,aEndPos,senseA,seqA,alignment,tempCalcA);
  getSegmentsFromBlock(&bLinks,bStartPos,bEndPos,senseB,seqB,alignment,tempCalcB);


  // compute the means and totals
  float aTotal = 0;
  for (map<long,float>::iterator it = aLinks.begin(); it != aLinks.end(); ++it){
    aTotal += it->second;
  }
  float aMean = aTotal / aLinks.size();
  if (aMean < 0.0){ aMean = 0.0; }
  float bTotal = 0;
  for (map<long,float>::iterator it = bLinks.begin(); it != bLinks.end(); ++it){
    bTotal += it->second;
  }
  float bMean = bTotal / bLinks.size();
  if (bMean < 0.0){ bMean = 0.0; }


  // pick the winner and add to the contig
  float penaltyConstant;
  if (aMean > bMean){ // seqA won
    if (aTotal==0){ penaltyConstant = 1; } // avoid divide-by-zero error
    else { penaltyConstant = 1.0 - bMean * aLinks.size() / aTotal; }
    fillInGaps(carrier, seqA, aStartPos, aEndPos, senseA, penaltyConstant, &aLinks, tempCalcA);
  } else if (aMean < bMean) { // seqB won
    if (bTotal==0){ penaltyConstant = 1; } // avoid divide-by-zero error
    else { penaltyConstant = 1.0 - aMean * bLinks.size() / bTotal; }
    fillInGaps(carrier, seqB, bStartPos, bEndPos, senseB, penaltyConstant, &bLinks, tempCalcB);
  } else { // there was a tie; this could be due to an edge-of-sequence issue
    // we are at the edge of A, so B is chosen
    if (bTotal==0){ penaltyConstant = 1; } // avoid divide-by-zero error
    else { penaltyConstant = 1.0 - aMean * bLinks.size() / bTotal; }
    if ( (aStartPos == -1 and aEndPos != seqA->size() and 
	  alignment->isGapped(0,seqB) and 
	  alignment->gapPairedAfter(0,seqB) == -1 ) or
	 (aEndPos == seqA->size() and aStartPos != -1 and
	  alignment->isGapped(seqB->size()-1,seqB) and
	  alignment->gapPairedAfter(bEndPos-1,seqB) == seqA->size()-1) ){
      fillInGaps(carrier, seqB, bStartPos, bEndPos, senseB, penaltyConstant, &bLinks, tempCalcB);
    }
    // we are at the edge of B, so A is chosen - or, it is just a tie so A is chosen
    else { fillInGaps(carrier, seqA, aStartPos, aEndPos, senseA, penaltyConstant, &aLinks, tempCalcA); }

  }
  delete tempCalcA;
  delete tempCalcB;
}



AssemblyJobHelperStrategies::CombineDataCarrier::CombineDataCarrier(){}
AssemblyJobHelperStrategies::CombineDataCarrier::CombineDataCarrier(long maxSize) : _maxSize(maxSize) {
  _seq = new char[_maxSize];
  _scores = new float[_maxSize];
  _links = new float[_maxSize];
  _seq[0] = '\0';
  _size = 0;
  _linkOffset = 0;
}
AssemblyJobHelperStrategies::CombineDataCarrier::~CombineDataCarrier(){
  delete [] _seq;
  delete [] _scores;
  delete [] _links;
}
void AssemblyJobHelperStrategies::CombineDataCarrier::addNuc(char nuc, float score){
  _seq[_size] = nuc;
  _scores[_size] = score;
  _linkOffset--;
  _size++;
  _seq[_size] = '\0';
}
void AssemblyJobHelperStrategies::CombineDataCarrier::addNuc(char nuc, float score, float link){
  _seq[_size] = nuc;
  _scores[_size] = score;
  _links[_size + _linkOffset] = link;
  _size++;
  _seq[_size] = '\0';
}
void AssemblyJobHelperStrategies::CombineDataCarrier::addLink(float link){
  _links[_size + _linkOffset] = link;
  _linkOffset++;
}
void AssemblyJobHelperStrategies::CombineDataCarrier::decrementLink(){
  _linkOffset--;
}
ScoredSeq* AssemblyJobHelperStrategies::CombineDataCarrier::getSeq(){
  return new ScoredSeqShallow(_seq, _scores, _links, _size);
}

bool AssemblyJobHelperStrategies::isPositionFinal(ScoredSeq * seqX, long posX, ScoredSeq * seqY, long posY){
  return isPositionFinal(seqX,posX) or isPositionFinal(seqY,posY);
}
bool AssemblyJobHelperStrategies::isPositionFinal(ScoredSeq * seq, long pos){
  return pos + 1 == seq->size();
}





#endif
